
/*! 
 * Business elements are aspects of a knowledge field: product, market, 
 * technology, organization, etc.
 * Business elements instances are a particular product, a particular market, 
 * etc.
 * 
 *  \author Sirio Capizzi
 *  \author Guido Fioretti
 *  \author Ruggero Rossi
 *  \copyright GNU Public License.
 */
public class BusinessElementInstance {
	
	
	private int whichAspect;  ///< Which aspect of a knowledge field this is an instance of.
	private boolean undefinedAspect;	///< Undefined aspects are created by  the partial intersection 
										///< of regular knowledge fields.
	private long instanceOfThisAspect;	///< The identification number of each single instance
	
	
	
	
	private static int last_serial_id = 0;	///< The value of this static 
	      								 	///< variable  is common to all.
											///<
	
	/**
	 * 
	 */
	private static synchronized void IncrementLastSerialID()
	{
		last_serial_id++;
	}
	
	/**
	 * 
	 * @return
	 */
	private static synchronized long getLastSerialID()
	{
		return last_serial_id;
	}
	
	/**
	 * 
	 */
	private void setID()
	{
		instanceOfThisAspect = new Long(getLastSerialID());
		IncrementLastSerialID();
	}
	
	
	
	
	
	
	/**
	 * 
	 * @return
	 */
	public long getInstanceOfAspect()
	{
		return instanceOfThisAspect;
	}

	/**
	 * An instance of an aspect of knowledge is created keeping track of:
	 * - Which aspect of knowledge it is an instance of (0 = first, 1 = second, ...); 
	 * - Whether it is undefined;
	 * - The ID of this particular instance of an aspect of a knowledge field.
	 *
	 * @param aBusinessElement
	 * @param isUndefined
	 */
	public BusinessElementInstance(int anAspect, boolean isUndefined) 
	{
	    whichAspect = anAspect;
	    undefinedAspect = isUndefined;
		setID();
	}

	/**
	 * Reports which aspect of knowledge this is an instance of.
	 * 
	 */
	public int getWhichAspect() {
		return whichAspect;
	}
	
	/**
	 * Reports whether the aspect of knowledge is undefined.
	 * 
	 */
	public boolean isAspectUndefined() {
		return undefinedAspect;
	}
}
