
import java.util.Arrays;


/*!
 * A knowledge field is the aggregation of business elements instances (e.g. 
 * a product, a market, a technology, etc.).
 * 
 *  \author Sirio Capizzi
 *  \author Guido Fioretti
 *  \author Ruggero Rossi
 *  \copyright GNU Public License.
 */
public class KnowledgeField {

	public BusinessElementInstance[] instances;	///< The instances of business 
												///< elements. Initially, this 
												///< vector is void.
	private int numBEI = Model.numBusinessElements;	///< Their number equals the 
													///< number of business 
													///< elements.
	private int n;	///< A counter for BEIs while they are being added to 
					///< 'instances'.
	
	// Identification of knowledge fields.
	private Long id;  ///< the identification number of each single instance
	private static long last_serial_id = 0;	///< the value of this static 
											///< variable is common to all 
											///< instances
	
	/**
	 * 
	 */
	private static synchronized void incrementLastSerialID()
	{
		last_serial_id++;
	}
	
	/**
	 * 
	 * @return
	 */
	private static synchronized long getLastSerialID()
	{
		return last_serial_id;
	}
	
	/**
	 * 
	 */
	private void setID()
	{
		id=new Long(getLastSerialID());
		incrementLastSerialID();
	}
	
	/**
	 * 
	 * @return
	 */
	public Long getID()
	{
		return id;
	}
	
	/**
	 *  Common Knowledge may need to change a knowledge field's ID if it 
	 *  discovers that it has the same BEIs of an existing one.
	 * @param newID
	 */
	public void changeID(long newID)
	{
		id = newID;
	}
			
	/**
	 * Creates a new knowledge field with its identification number.
	 * At the beginning, with no business element instances.
	 * BEIs are added at a later step using the method 'addBEI' below.
	 */
	public KnowledgeField()
	{
		n = 0;
		instances=new BusinessElementInstance[numBEI];
		setID();
	}
	
	/**
	 * 
	 * @param bei
	 */
	public void addBEI(BusinessElementInstance bei) {
		instances[n++] = bei;
	}
	

	
	/**
	 * Substitutes a business element instance.
	 * The be_i-th business element instance takes the value bei.
	 * 
	 * @param be_i
	 * @param bei
	 */
	void substituteBEI(int be_i, BusinessElementInstance bei) {
System.out.println("E' dentro substituteBEI");
		instances[be_i] = bei;
	}
	
	/**
	 * Returns true if at least one of the business elements instances of this 
	 * knowledgefield is undefined.
	 *
	 * @return
	 */
	public boolean entailsUndefinedBEI() {
System.out.println("sono dentro entailsUndefinedBEI");
		boolean entails = false;
		for (int i=0; i<numBEI; i++) 
			if (instances[i].isAspectUndefined()) 
				entails = true;
		return entails;
	}
	
	
	
	/**
	 * This method should only be called on a knowledge field entailing 
	 * undefined elements.
	 * It creates a new knowledge field by composing a knowledge field
	 * where some business elements are undefined and a knowledge field
	 * where all business elements should be regular. So if one knowledge field
	 * is [UNDEFINED,I2,UNDEFINED] and the other knowledge field is [I3,I4,I5],
	 * the ensuing knowledge field will be [I3,I2,I5].
	 * This method is called on knowledge fields entailing undefined business elements instances;
	 * the knowledge field entailing only regular business elements instances is passed to it.
	 * However, if the filler kf has undefined aspects where this also has, then an undefined kf results.
	 * @param filler
	 * @return
	 */
	public KnowledgeField instantiateWith(KnowledgeField filler) {
System.out.println("E' dentro instantiateWith");
		/*if (filler.entailsUndefinedBEI())
			System.out.println("instantiateWith viene chiamato con un kf undefined");*/
/*IMPORTANTE:
Se le righe precedenti vengono commentate, la NullPointerException non viene piu' sulla riga 145
ma sulla riga 157. Quindi il problema e' su filler.
INOLTRE: Se queste righe non vengono commentate non viene fuori la scritta "sono dentro entailsUndefinedBEI" 
(r. 145), che invece viene fuori se vengono commentate (dalla r. 152). Altro indizio che si tratta di un 
problema con filler.*/
		if (this.entailsUndefinedBEI()) {
			KnowledgeField kfr = this.cloneWithNewID();
			for (int i=0; i<Model.numBusinessElements; i++)	{
				if (kfr.instances[i].isAspectUndefined())
					kfr.substituteBEI(i, filler.instances[i]);  // undefined 
														//BEI: take filler's BEI
			}
			return kfr;
		}
		else
			return this;
	}
		
	/**
	 * Returns the extent to which kf overlaps on this knowledge field.
	 * This is the ratio between the number of common BE instances between kf 
	 * and this knowledge field, and the number of BE instances in this knowledge field.
	 * It is called among else by Dempster-Shafer, which may deal with undefined BEIs.
	 *
	 * @param kf
	 * @return
	 */
  	public double intersectionValue(KnowledgeField kf) {
		double intValue;
		// Intersection is 1 if this knowledge field is identical to kf.
		if (kf.getID() == id) {
			intValue = 1.0;
		}
		else {
			BusinessElementInstance[] ba1=instances;
			BusinessElementInstance[] ba2=kf.instances;
			// BEIs intersect either if they coincide, or if at least one of 
			// them is undefined.
			int inter=0;
			for (int i=0; i<numBEI; i++) {		
					if (ba1[i].getInstanceOfAspect() == ba2[i].getInstanceOfAspect() || ba1[i].isAspectUndefined() 
							|| ba2[i].isAspectUndefined()) { 
					inter++;								
				}
			}
			intValue = (double)inter/(double)numBEI;
			if (intValue > 1.0)
				throw new IllegalStateException("Intersection greater than "
						+ "one!");
		}
		return intValue;
	}
	
	/**
	 * Returns true if this fields and the calling one intersect.
	 * It is called among else by Dempster-Shafer, which may deal with undefined aspect instances.
	 *
	 * @param kf
	 * @return
	 */
	public boolean intersects(KnowledgeField kf) {
System.out.println("intersects viene chiamato");
System.out.println("kf1 contiene " + kf.instances[0].getInstanceOfAspect() + " e " + kf.instances[1].getInstanceOfAspect());
System.out.println("kf2 contiene " + this.instances[0].getInstanceOfAspect() + " e " + this.instances[1].getInstanceOfAspect());
		boolean answer = false;
		BusinessElementInstance[] ba1=instances;
		BusinessElementInstance[] ba2=kf.instances;
		for (int i=0; i<numBEI; i++) {
			if (ba1[i].isAspectUndefined() == true && ba2[i].isAspectUndefined() == false
					|| ba1[i].isAspectUndefined() == false && ba2[i].isAspectUndefined() == true
					|| ba1[i].isAspectUndefined() == false && ba2[i].isAspectUndefined() == false && 
					ba1[i].getInstanceOfAspect() == ba2[i].getInstanceOfAspect())
					answer = true;
		}
System.out.println("intersects dice: " + answer);
		return answer;
	}

	/**
	 * Returns true if this field is equivalent to the calling one.
	 * Knowledge fields that have undefined BEIs may be equivalent even if they 
	 * have different IDs. It is called by Dempster-Shafer, which may deal with undefined BEIs.
	 *
	 * @param kf
	 * @return
	 */
	public boolean isEquivalent(KnowledgeField kf) {
		boolean answer = true;
		BusinessElementInstance[] ba1=instances;
		BusinessElementInstance[] ba2=kf.instances;
		for (int i=0; i<numBEI; i++) {
			if (ba1[i].isAspectUndefined() == true && ba2[i].isAspectUndefined() == false
					|| ba1[i].isAspectUndefined() == false && 
					ba2[i].isAspectUndefined() == true
					|| ba1[i].isAspectUndefined() == false && 
					ba2[i].isAspectUndefined() == false && 
					ba1[i].getInstanceOfAspect() != ba2[i].getInstanceOfAspect())
				answer = false;
		}
		return answer;
	}

	/**
	 * Returns true if this field includes the calling one.
	 * Returns "true" if this knowledge field has a regular aspect where the calling field has an undefined one.
	 * It is called by Dempster-Shafer, which may deal with undefined BEIs.
	 *
	 * @param kf
	 * @return
	 */
	public boolean includes(KnowledgeField kf) {
		boolean answer = true;
		BusinessElementInstance[] ba1=instances;
		BusinessElementInstance[] ba2=kf.instances;
		for (int i=0; i<numBEI; i++) {
			if (ba1[i].isAspectUndefined() == false && 
					ba2[i].isAspectUndefined() == false && 
					ba1[i].getInstanceOfAspect() != ba2[i].getInstanceOfAspect() 
					|| ba1[i].isAspectUndefined() == true && 
					ba2[i].isAspectUndefined() == false)
				answer = false;
		}
		return answer;
	}


		
	/**
	 * Returns one business element instance of this knowledge field.
	 * 
	 * @param i
	 * @return
	 */
	public BusinessElementInstance getInstance(int i) {
		return instances[i];
	}
	
	/**
	 * Creates a new knowledge field, identical to this one except for its 
	 * identification number.
	 * Called by Experiential Exploration. Used by Vicarious Exploration 
	 * through instantiateWith.
	 *
	 * @return
	 */
	public KnowledgeField cloneWithNewID() {
		KnowledgeField clone = new KnowledgeField();	// it automatically gets
														//a new ID
		clone.n = numBEI;  // BEIs are not added one-by-one to the clone
		clone.instances = new BusinessElementInstance[numBEI];		
		System.arraycopy(instances, 0, clone.instances, 0, numBEI);
		return clone;
	}
}
